"""Widgets that visualise volume images in .nii files."""
import inspect
import os.path

import ipywidgets as widgets
import matplotlib.pyplot as plt
import nibabel as nib
import numpy as np
import scipy.ndimage
from ipywidgets import IntSlider, fixed, interact

from .colormaps import get_cmap_dropdown


class NiftiWidget:
    """Turn .nii files into interactive plots using ipywidgets.

    Args
    ----
        filename : str
                The path to your ``.nii`` file. Can be a string, or a
                ``PosixPath`` from python3's pathlib.
    """

    def __init__(self, filename):
        """
        Turn .nii files into interactive plots using ipywidgets.

        Args
        ----
            filename : str
                    The path to your ``.nii`` file. Can be a string, or a
                    ``PosixPath`` from python3's pathlib.
        """
        if hasattr(filename, 'get_data'):
            self.data = filename
        else:
            filename = str(filename)
            if not os.path.isfile(filename):
                raise OSError('File ' + filename + ' not found.')

            # load data in advance
            # this ensures once the widget is created that the file is of a
            # format readable by nibabel
            self.data = nib.load(str(filename))

        # initialise where the image handles will go
        self.image_handles = None

    def nifti_plotter(self, plotting_func=None, colormap=None, figsize=(15, 5),
                      **kwargs):
        """
        Plot volumetric data.

        Args
        ----
            plotting_func : function
                    A plotting function for .nii files, most likely
            mask_background : bool
                    Whether the background should be masked (set to NA).
                    This parameter only works in conjunction with the default
                    plotting function (`plotting_func=None`). It finds clusters
                    of values that round to zero and somewhere touch the edges
                    of the image. These are set to NA. If you think you are
                    missing data in your image, set this False.
            colormap : str | list
                    The matplotlib colormap that should be applied to the data.
                    By default, the widget will allow you to pick from all that
                    are available, but you can pass a string to fix the
                    colormap or a list of strings to offer the user a few
                    options.
            figsize : tup
                    The figure height and width for matplotlib, in inches.

            If you are providing a custom plot function, any kwargs you provide
            to nifti_plotter will be passed to that function.

        """
        kwargs['colormap'] = get_cmap_dropdown(colormap)
        kwargs['figsize'] = fixed(figsize)

        if plotting_func is None:
            self._default_plotter(**kwargs)
        else:
            self._custom_plotter(plotting_func, **kwargs)

    def _default_plotter(self, mask_background=False, **kwargs):
        """Plot three orthogonal views.

        This is called by nifti_plotter, you shouldn't call it directly.
        """
        plt.gcf().clear()
        plt.ioff()  # disable interactive mode

        data_array = self.data.get_data()

        if not ((data_array.ndim == 3) or (data_array.ndim == 4)):
            raise ValueError('Input image should be 3D or 4D')

        # mask the background
        if mask_background:
            # TODO: add the ability to pass 'mne' to use a default brain mask
            # TODO: split this out into a different function
            if data_array.ndim == 3:
                labels, n_labels = scipy.ndimage.measurements.label(
                                            (np.round(data_array) == 0))
            else:  # 4D
                labels, n_labels = scipy.ndimage.measurements.label(
                    (np.round(data_array).max(axis=3) == 0)
                )

            mask_labels = [lab for lab in range(1, n_labels+1)
                           if (np.any(labels[[0, -1], :, :] == lab) |
                               np.any(labels[:, [0, -1], :] == lab) |
                               np.any(labels[:, :, [0, -1]] == lab))]

            if data_array.ndim == 3:
                data_array = np.ma.masked_where(
                    np.isin(labels, mask_labels), data_array)
            else:
                data_array = np.ma.masked_where(
                    np.broadcast_to(
                        np.isin(labels, mask_labels)[:, :, :, np.newaxis],
                        data_array.shape
                    ),
                    data_array
                )

        # init sliders for the various dimensions
        for dim, label in enumerate(['x', 'y', 'z']):
            if label not in kwargs.keys():
                kwargs[label] = IntSlider(
                    value=(data_array.shape[dim] - 1)/2,
                    min=0, max=data_array.shape[dim] - 1,
                    continuous_update=False
                )

        if (data_array.ndim == 3) or (data_array.shape[3] == 1):
            kwargs['t'] = fixed(None)  # time is fixed
        else:
            kwargs['t'] = IntSlider(
                value=0, min=0, max=data_array.shape[3] - 1,
                continuous_update=False
            )

        widgets.interact(self._plot_slices, data=fixed(data_array), **kwargs)

        plt.close()  # clear plot
        plt.ion()  # return to interactive state

    def _plot_slices(self, data, x, y, z, t,
                     colormap='viridis', figsize=(15, 5)):
        """
        Plot x,y,z slices.

        This function is called by _default_plotter
        """
        fresh = self.image_handles is None
        if fresh:
            self._init_figure(data, colormap, figsize)

        coords = [x, y, z]

        # add plot titles to the subplots
        views = ['Sagittal', 'Coronal', 'Axial']
        for i, ax in enumerate(self.fig.axes):
            ax.set_title(views[i])

        for ii, imh in enumerate(self.image_handles):

            slice_obj = 3 * [slice(None)]

            if data.ndim == 4:
                slice_obj.append(t)

            slice_obj[ii] = coords[ii]

            # update the image
            imh.set_data(
                np.flipud(np.rot90(data[slice_obj], k=1))
                if views[ii] != 'Sagittal' else
                np.fliplr(np.flipud(np.rot90(data[slice_obj], k=1)))
            )

            # draw guides to show selected coordinates
            guide_positions = [val for jj, val in enumerate(coords)
                               if jj != ii]
            imh.axes.lines[0].set_xdata(2*[guide_positions[0]])
            imh.axes.lines[1].set_ydata(2*[guide_positions[1]])

            imh.set_cmap(colormap)

        if not fresh:
            return self.fig

    def _init_figure(self, data, colormap, figsize):
        # init an empty list
        self.image_handles = []
        # open the figure
        self.fig, axes = plt.subplots(1, 3, figsize=figsize)

        for ii, ax in enumerate(axes):

            ax.set_facecolor('black')

            ax.tick_params(
                axis='both', which='both', bottom='off', top='off',
                labelbottom='off', right='off', left='off', labelleft='off'
                )
            # fix the axis limits
            axis_limits = [limit for jj, limit in enumerate(data.shape[:3])
                           if jj != ii]
            ax.set_xlim(0, axis_limits[0])
            ax.set_ylim(0, axis_limits[1])

            img = np.zeros(axis_limits[::-1])
            # img[1] = data_max
            im = ax.imshow(img, cmap=colormap,
                           vmin=data.min(), vmax=data.max())
            # add "cross hair"
            ax.axvline(x=0, color='gray', alpha=0.8)
            ax.axhline(y=0, color='gray', alpha=0.8)
            # append to image handles
            self.image_handles.append(im)
        # plt.show()

    def _custom_plotter(self, plotting_func, **kwargs):
        """Collect data and start interactive widget for custom plot."""
        self.plotting_func = plotting_func
        plt.gcf().clear()
        plt.ioff()

        # XYZ Sliders if plot supports it and user didn't provide any:
        if ('cut_coords' in inspect.getargspec(self.plotting_func)[0]
                and 'cut_coords' not in kwargs.keys()):
            for label in ['x', 'y', 'z']:
                if label not in kwargs.keys():
                    # cut_coords should be given in MNI coordinates
                    kwargs[label] = IntSlider(value=0, min=-90, max=90,
                                              continuous_update=False)

        # Create the widget:
        interact(self._custom_plot_wrapper, data=fixed(self.data), **kwargs)
        plt.ion()

    def _custom_plot_wrapper(self, data, **kwargs):
        """Wrap a custom function."""
        # start the figure
        fig = plt.figure(figsize=kwargs.pop('figsize', None))

        # The following should provide a colormap option to most plots:
        if 'colormap' in kwargs.keys():
            if 'cmap' in inspect.getargspec(self.plotting_func)[0]:
                # if cmap is valid argument to plot func, rename colormap
                kwargs['cmap'] = kwargs.pop('colormap')
            else:
                # if cmap is not valid for plot func, try and coerce it
                plt.set_cmap(kwargs.pop('colormap'))

        # reconstruct manually added x-y-z-sliders:
        if ('cut_coords' in inspect.getargspec(self.plotting_func)[0]
                and 'x' in kwargs.keys()):

            # add the x-y-z as cut_coords
            if ('display_mode' not in kwargs.keys()
                or not any([label in kwargs['display_mode']
                            for label in ['x', 'y', 'z']])):
                # If no xyz combination of display modes was requested:
                kwargs['cut_coords'] = [kwargs[label]
                                        for label in ['x', 'y', 'z']]
            else:
                kwargs['cut_coords'] = [kwargs[label]
                                        for label in ['x', 'y', 'z']
                                        if label in kwargs['display_mode']]
            # remove x-y-z from kwargs
            [kwargs.pop(label, None) for label in ['x', 'y', 'z']]

        # Actually plot the image
        self.plotting_func(data, figure=fig, **kwargs)
        plt.show()


class VolumeWidget:
    """Turn .nii files into interactive plots using ipywidgets.

    Args
    ----
        filename : str
                The path to your ``.nii`` file. Can be a string, or a
                ``PosixPath`` from python3's pathlib.
    """

    def __init__(self, filename):
        """
        Turn .nii files into interactive plots using ipywidgets.

        Args
        ----
            filename : str
                    The path to your ``.nii`` file. Can be a string, or a
                    ``PosixPath`` from python3's pathlib.
        """
        if hasattr(filename, 'get_data'):
            self.data = filename
        else:
            filename = str(filename)
            if not os.path.isfile(filename):
                raise OSError('File ' + filename + ' not found.')

            # load data in advance
            # this ensures once the widget is created that the file is of a
            # format readable by nibabel
            self.data = nib.load(str(filename))

        # initialise where the image handles will go
        self.image_handles = None

        # initialise the control components of this widget

    def plot():
        pass



"""
Widgets to visualise neuroimaging data.

For volume images, try import NiftiWidget.
For surface images, try SurfaceWidget.
"""
from .version import __version__  # noqa: F401
from .exampledata import exampleatlas, examplezmap, examplet1  # noqa: F401
from .niwidget_volume import NiftiWidget  # noqa: F401
from .niwidget_surface import SurfaceWidget  # noqa: F401
from .streamlines import StreamlineWidget  # noqa: F401



"""Example files for use with niwidgets."""
__all__ = ['exampleatlas', 'examplezmap', 'examplet1',
           'examplesurface', 'exampleoverlays', 'streamlines']

import os.path

root_dir = os.path.join(os.path.dirname(__file__), 'data')

exampleatlas = os.path.join(root_dir, 'cc400_roi_atlas.nii')
examplezmap = os.path.join(root_dir, 'cognitive control_pFgA_z.nii.gz')
examplet1 = os.path.join(root_dir, 'T1.nii.gz')

surface_dir = os.path.join(root_dir, 'example_surfaces')

examplesurface = os.path.join(surface_dir, 'lh.inflated')

exampleoverlays = {
    key: os.path.join(surface_dir, f)
    for key, f in zip(['Area', 'Curvature', 'Thickness', 'Annotation'],
                      ('lh.area', 'lh.curv', 'lh.thickness', 'lh.aparc.annot'))
}

streamlines = os.path.join(root_dir, 'streamlines.trk')



"""A widget for surface-warped neuroimaging data."""
from __future__ import print_function

import os
from collections import defaultdict
from xml.parsers.expat import ExpatError

import ipyvolume.pylab as p3
import matplotlib.pyplot as plt
import nibabel as nb
import numpy as np
from IPython.display import display
from ipywidgets import Dropdown, fixed, interact

from .colormaps import get_cmap_dropdown


def _check_file(file):
    """Check if file exists and if it's valid"""
    if isinstance(file, nb.gifti.gifti.GiftiImage):
        return file
    elif os.path.isfile(str(file)):
        return str(file)
    else:
        raise ValueError('The argument meshfile needs to either be '
                         'a nibabel Gifti image or an existing file, '
                         'but ' + str(file) + ' was provided.')


def _get_name(file):
    """Get a name for the supplied file / giftiimage"""
    if isinstance(file, nb.gifti.gifti.GiftiImage):
        if file.get_filename():
            return os.path.basename(file.get_filename())
        else:
            return str(file)
    elif os.path.isfile(str(file)):
        return os.path.basename(str(file))


class SurfaceWidget:
    """Interact with brain surfaces right in the notebook."""

    def __init__(self, meshfile, overlayfiles=()):
        """Create a surface widget.

        This widget takes in surface data, in the form of gifti files or
        freesurfer files, and displays them interactively.

        meshfile : str, Path, nb.gifti.gifti.GiftiImage
            A file containing the the surface information. It can either be
            a Freesurfer file (i.e. lh.pial) or a .gii mesh file. A loaded
            GiftiImage (using nibabel) will also work.

        overlayfiles : tuple, dict, str, nb.gifti.gifti.GiftiImage
            Data you'd like to overlay on the 3D mesh. This can be a tuple of
            files, a dictionary (in which case the keys of the dict are used
            as options in a dropdown menu), or a single file name / loaded
            GiftiImage. Possible file formats: .annot, .thickness, .curv, .sulc
            or .gii.

        """
        # check meshfile is a valid argument
        self.meshfile = _check_file(meshfile)

        # make sure overlayfiles is a dictionary
        if isinstance(overlayfiles, dict):
            self.overlayfiles = overlayfiles
        elif isinstance(overlayfiles, (list, tuple)):
            self.overlayfiles = {
                _get_name(file): _check_file(file)
                for file in overlayfiles
            }
        else:
            self.overlayfiles = {_get_name(overlayfiles):
                                 _check_file(overlayfiles)}

        self.fig = None

    def _init_figure(self, x, y, z, triangles, figsize, figlims):
        """
        Initialize the figure by plotting the surface without any overlay.

        x: x coordinates of vertices (V x 1 numpy array)
        y: y coordinates of vertices (V x 1 numpy array)
        z: z coordinates of vertices (V x 1 numpy array)
        triangles: triangle specifications (T x 3 numpy array, where
        T=#triangles)
        figsize: 2x1 list of integers
        figlims: 3x2 list of integers
        """
        self.fig = p3.figure(width=figsize[0], height=figsize[1])
        self.fig.camera_fov = 1
        self.fig.style = {'axes': {'color': 'black',
                                   'label': {'color': 'black'},
                                   'ticklabel': {'color': 'black'},
                                   'visible': False},
                          'background-color': 'white',
                          'box': {'visible': False}}
        self.fig.xlim = (figlims[0][0], figlims[0][1])
        self.fig.ylim = (figlims[1][0], figlims[1][1])
        self.fig.zlim = (figlims[2][0], figlims[2][1])

        # draw the tetrahedron
        p3.plot_trisurf(x, y, z, triangles=triangles,
                        color=np.ones((len(x), 3)))

    def _plot_surface(self, x, y, z, triangles,
                      overlays=None, frame=0,
                      colormap='summer', figsize=np.array([600, 600]),
                      figlims=np.array(3 * [[-100, 100]])):
        """
        Visualize/update the overlay.

        This function changes the color associated with mesh vertices.

        overlays: V x F numpy array, where each column corresponds to a
                  different overlay. F=#frames or #timepoints.

        """
        if self.fig is None:
            self._init_figure(x, y, z, triangles, figsize, figlims)
        # overlays is a 2D matrix
        # with 2nd dimension corresponding to (time) frame
        my_color = plt.cm.get_cmap(colormap)
        if overlays[frame] is not None:
            # activation = overlays[:, frame]
            activation = overlays[frame]
            if max(activation) - min(activation) > 0:
                colors = my_color((activation - min(activation))
                                  / (max(activation) - min(activation)))
            else:
                colors = my_color((activation - min(activation)))
            self.fig.meshes[0].color = colors[:, :3]

    def zmask(surf, mask):
        """
        Masks out vertices with intensity=0 from overlay.

        Also returns masked-out vertices.

        Parameters
        ----------
        surf: gifti object
            already loaded gifti object of target surface
        mask: gifti object
            already loaded gifti object of overlay with zeroes
            at vertices of no interest (e.g. medial wall)

        Returns
        -------
        mask_keep : np.ndarray
            Boolean array of what to show.
        mask_kill : np.ndarray
            Boolean array of what to hide.

        """
        ikill = np.flatnonzero(mask.darrays[0].data == 0)

        # create empty arrays matching surface mesh dimentions
        mask_kill = np.zeros([surf.darrays[1].data.shape[0]], dtype=bool)

        for ii, row in enumerate(surf.darrays[1].data):
            for item in row:
                if item in ikill:
                    mask_kill[ii] = True

        return ~mask_kill, mask_kill

    def surface_plotter(self, colormap=None, figsize=np.array([600, 600]),
                        figlims=np.array(3 * [[-100, 100]]),
                        show_zeroes=True, **kwargs):
        """Visualise a surface mesh (with overlay) inside notebooks.

        This method displays the surface widget.

        Parameters
        ----------
        surface : str, gifti object
            Path to surface file in gifti or FS surface format or an
            already loaded gifti object of surface
        overlay : str, gifti object
            Path to overlay file in gifti or FS annot or anatomical
            (.curv,.sulc,.thickness) format or an already loaded
            gifti object of overlay, default None
        colormap : string
            A matplotlib colormap, default summer
        figsize : ndarray
            Size of the figure to display, default [600,600]
        figlims : ndarray
            x,y and z limits of the axes, default
            [[-100,100],[-100,100],[-100,100]]
        show_zeroes : bool
            Display vertices with intensity = 0, default True

        """
        kwargs['colormap'] = get_cmap_dropdown(colormap)
        kwargs['figsize'] = fixed(figsize)
        kwargs['figlims'] = fixed(figlims)

        if isinstance(self.meshfile, str):
            # if mesh has not been loaded before, load it
            if os.path.splitext(self.meshfile)[1].lower() == '.gii':
                # load gifti file
                try:
                    self.meshfile = nb.load(self.meshfile)
                    x, y, z = self.meshfile.darrays[0].data.T
                    vertex_edges = self.meshfile.darrays[1].data
                except ExpatError:
                    raise ValueError(
                        'The file {} could not be read. '.format(self.meshfile)
                        + 'Please provide a valid gifti file.')
            else:
                # load freesurfer file
                fsgeometry = nb.freesurfer.read_geometry(self.meshfile)
                x, y, z = fsgeometry[0].T
                vertex_edges = fsgeometry[1]

        elif isinstance(self.meshfile, nb.gifti.gifti.GiftiImage):
            # if mesh has been loaded as a GiftiImage, format the data
            x, y, z = self.meshfile.darrays[0].data.T
            vertex_edges = self.meshfile.darrays[1].data

        overlays = defaultdict(lambda: None)
        if len(self.overlayfiles) > 0:

            for key, overlayfile in self.overlayfiles.items():

                file_ext = os.path.splitext(overlayfile)[1].lower()

                if isinstance(overlayfile, nb.gifti.gifti.GiftiImage):
                    overlays[key] = overlayfile.darrays[0].data
                else:
                    file_ext = os.path.splitext(overlayfile)[1].lower()

                    if file_ext == '.gii':
                        try:
                            overlay = nb.load(overlayfile)
                            overlays[key] = overlay.darrays[0].data
                        except ExpatError:
                            raise ValueError(
                                'The file {} could not be read. '
                                .format(overlayfile)
                                + 'Please provide a valid gifti file.')

                    elif (file_ext in ('.annot', '')):
                        annot = nb.freesurfer.read_annot(overlayfile)
                        overlays[key] = annot[0]

                    elif (file_ext in ('.curv', '.thickness', '.sulc')):
                        overlays[key] = nb.freesurfer.read_morph_data(
                            overlayfile)

                    if not show_zeroes:
                        pass

        kwargs['triangles'] = fixed(vertex_edges)
        kwargs['x'] = fixed(x)
        kwargs['y'] = fixed(y)
        kwargs['z'] = fixed(z)
        kwargs['overlays'] = fixed(overlays)

        if len(self.overlayfiles) < 2:
            frame = fixed(None)
        else:
            frame = Dropdown(
                options=list(self.overlayfiles.keys()),
                value=list(self.overlayfiles.keys())[0],
                description='Overlay:'
            )

        interact(self._plot_surface, frame=frame, **kwargs)
        display(self.fig)



from __future__ import print_function

import os

import ipyvolume as ipv
import nibabel as nib
import numpy as np
from ipywidgets import fixed, interact, widgets


def length(x):
    """Returns the sum of euclidean distances between neighboring points"""
    return np.sum(np.sqrt(np.sum((x[:-1, :] - x[1:, :]) *
                                 (x[:-1, :] - x[1:, :]),
                                 axis=1)))


def color(x):
    """Returns an approximation for color of line based on its endpoints"""
    dirvec = (x[0, :] - x[-1, :])
    return (dirvec/(np.sqrt(np.sum(dirvec*dirvec, axis=-1)))).dot(np.eye(3))


class StreamlineWidget:
    """
    Turns nibabel track files into interactive plots using ipyvolume.

    Color for each line is rendered as a function of the endpoints of the
    streamline. Currently, this resolves to red for left-right, green for
    anterior-posterior, and blue for inferior-superior.

    Args
    ----
        filename : str, pathlib.Path
                The path to your ``.trk`` file. Can be a string, or a
                ``PosixPath`` from python3's pathlib.
        streamlines : a nibabel streamline object
                An streamlines attribute of an object loaded by
                nibabel.streamlines.load
    """

    def __init__(self, filename=None, streamlines=None):

        if filename:
            filename = str(filename)
            if not os.path.isfile(filename):
                # for Python3 should have FileNotFoundError here
                raise IOError('file {} not found'.format(filename))

            if not nib.streamlines.is_supported(filename):
                raise ValueError(('File {0} is not a streamline file supported'
                                  ' by nibabel').format(filename))

            # load data in advance
            self.streamlines = nib.streamlines.load(filename).streamlines
        elif streamlines:
            self.streamlines = streamlines
        else:
            raise ValueError('One of filename or streamlines must be specified'
                             )
        self.lines2use_ = None

    def plot(self, display_fraction=0.1, **kwargs):
        """
        This is the main method for this widget.

        Args
        ----
            display_fraction : float
                    The fraction of streamlines to show
            percentile : int
                    The initial number of streamlines to show using a
                    percentile of length distribution
            width : int
                    The width of the figure
            height : int
                    The height of the figure
        """

        if display_fraction is not None and (display_fraction > 1 or
                                             display_fraction <= 0):
            raise ValueError('proportion_to_display is a float between 0 and 1'
                             ' (0 excluded) or None')

        N = len(self.streamlines)
        num_streamlines = int(display_fraction * N)
        indices = np.random.permutation(N)[:num_streamlines]
        self.lines2use_ = self.streamlines[indices]
        self._default_plotter(**kwargs)

    def _create_mesh(self, indices2use=None):
        if indices2use is None:
            lines2use = self.lines2use_
            local_colors = self.colors
        else:
            lines2use = self.lines2use_[indices2use]
            local_colors = self.colors[indices2use]
        x, y, z = np.concatenate(lines2use).T

        # will contain indices to the verties, [0, 1, 1, 2, 2, 3, 3, 4, 4, 5..]
        indices = np.zeros(np.sum((len(line) - 1) * 2 for
                                  line in lines2use), dtype=np.uint32)
        colors = np.zeros((len(x), 3), dtype=np.float32)

        vertex_offset = 0
        line_offset = 0
        line_pointers = []
        # if we have a line of 4 vertices, we need to add the indices:
        #  offset + [0, 1, 1, 2, 2, 3]
        # so we have approx 2x the number of indices compared to vertices
        for idx, line in enumerate(lines2use):
            line_length = len(line)
            # repeat all but the start and end vertex
            line_indices = np.repeat(np.arange(vertex_offset,
                                               vertex_offset + line_length,
                                               dtype=indices.dtype), 2)[1:-1]
            indices[line_offset:
                    line_offset + line_length * 2 - 2] = line_indices
            line_pointers.append([line_offset, line_length, line_indices])
            colors[vertex_offset:
                   vertex_offset + line_length] = local_colors[idx]
            line_offset += line_length * 2 - 2
            vertex_offset += line_length
        return x, y, z, indices, colors, line_pointers

    def _default_plotter(self, **kwargs):
        """
        Basic plot function to be used if no custom function is specified.

        This is called by plot, you shouldn't call it directly.
        """

        self.lengths = np.array([length(x) for x in self.lines2use_])
        if not('grayscale' in kwargs and kwargs['grayscale']):
            self.colors = np.array([color(x) for x in self.lines2use_])
        else:
            self.colors = np.zeros((len(self.lines2use_), 3), dtype=np.float16)
            self.colors[:] = [0.5, 0.5, 0.5]
        self.state = {'threshold': 0, 'indices': []}

        width = 600
        height = 600
        perc = 80
        if 'width' in kwargs:
            width = kwargs['width']
        if 'height' in kwargs:
            height = kwargs['height']
        if 'percentile' in kwargs:
            perc = kwargs['percentile']

        ipv.clear()
        fig = ipv.figure(width=width, height=height)
        self.state['fig'] = fig

        with fig.hold_sync():
            x, y, z, indices, colors, self.line_pointers = self._create_mesh()
            limits = np.array([min([x.min(), y.min(), z.min()]),
                               max([x.max(), y.max(), z.max()])])
            mesh = ipv.Mesh(x=x, y=y, z=z, lines=indices, color=colors)
            fig.meshes = [mesh]
            if 'style' not in kwargs:
                fig.style = {'axes': {'color': 'black',
                                      'label': {'color': 'black'},
                                      'ticklabel': {'color': 'black'},
                                      'visible': False},
                             'background-color': 'white',
                             'box': {'visible': False}}
            else:
                fig.style = kwargs['style']
            ipv.pylab._grow_limits(limits, limits, limits)
            fig.camera_fov = 1
        ipv.show()

        interact(self._plot_lines, state=fixed(self.state),
                 threshold=widgets.FloatSlider(
                     value=np.percentile(self.lengths, perc),
                     min=self.lengths.min() - 1,
                     max=self.lengths.max() - 1,
                     continuous_update=False))

    def _plot_lines(self, state, threshold):
        """
        Plots streamlines

        This function is called by _default_plotter
        """
        if threshold < state['threshold']:
            # when threshold is reduced, increase the number of lines
            state['indices'] = np.where(self.lengths > threshold)[0]
            with state['fig'].hold_sync():
                mesh = state['fig'].meshes[0]
                copy = mesh.lines.copy()
                for idx in state['indices']:
                    (line_offset, line_length,
                     line_indices) = self.line_pointers[idx]
                    copy[line_offset:line_offset + line_length * 2 - 2] = \
                        line_indices
                mesh.lines = copy
                mesh.send_state('lines')
        else:
            # when threshold is increased, decrease the number of lines
            indices = np.where(self.lengths <= threshold)[0]
            with state['fig'].hold_sync():
                mesh = state['fig'].meshes[0]
                copy = mesh.lines.copy()
                for idx in indices:
                    (line_offset, line_length,
                     line_indices) = self.line_pointers[idx]
                    copy[line_offset:line_offset + line_length * 2 - 2] = 0
                mesh.lines = copy
                mesh.send_state('lines')
            state['indices'] = np.where(self.lengths > threshold)[0]
        state['threshold'] = threshold



"""Installer."""
import os.path
# To use a consistent encoding
from codecs import open

from setuptools import setup

here = os.path.dirname(os.path.abspath(__file__))
ver_file = os.path.join(here, 'niwidgets', 'version.py')

with open(ver_file) as f:
    exec(f.read(), globals(), locals())

version = locals()['__version__']

blurb = 'A package that provides ipywidgets for standard neuroimaging plotting'

readme = (open('README.md', 'r').read()
          if os.path.isfile('README.md')
          else blurb)

setup(
    name='niwidgets',
    version=version,
    description=blurb,
    long_description=readme,
    long_description_content_type="text/markdown",
    url='https://github.com/nipy/niwidgets',
    download_url='https://github.com/nipy/niwidgets/archive/' +
        version + '.tar.gz',
    # Author details
    author='Bjoern Soergel & Jan Freyberg',
    author_email='jan.freyberg@gmail.com',
    packages=['niwidgets'],
    keywords=['widgets', 'neuroimaging'],
    install_requires=['ipywidgets', 'nibabel', 'ipyvolume',
                      'matplotlib', 'numpy', 'scipy'],
    # Include the template file
    package_data={
        '': ['data/*nii*',
             'data/*.trk',
             'data/examples_surfaces/lh.*',
             'data/examples_surfaces/*.ctab']
    },
    tests_require=[
        'pytest'
    ],
    setup_requires=[
        'pytest-runner'
    ]
)



import ipywidgets
from matplotlib import pyplot as plt


def get_cmap_dropdown(colormap):
    # set default colormap options & add them to the kwargs
    if colormap is None:
        options = (sorted(m for m in plt.cm.datad if not m.endswith("_r")))
        return ipywidgets.Dropdown(
            options=options, value='summer', description='Colormap:'
        )
    elif isinstance(colormap, (list, tuple)):
        return ipywidgets.Dropdown(
            options=colormap, value=colormap[0], description='Colormap:'
        )
    elif isinstance(colormap, str):
        return ipywidgets.fixed(colormap)
    else:
        raise ValueError('The colormap must be either a valid string, a list, '
                         'or None.')
